/*
    This file is part of Posixcpp library.

    Posixcpp library is free software: you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public License
    as published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    Posixcpp library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with Posixcpp library. If not, see
    <http://www.gnu.org/licenses/>.
*/

#ifndef POSIX_IO_HPP_INCLUDED
#define POSIX_IO_HPP_INCLUDED

#include <unistd.h>
#include <fcntl.h>
#include <poll.h>

#include <cstdlib>

#include "posixcpp.hpp"
#include "posixcpp_bytearray.hpp"
#include "posixcpp_flags.hpp"
#include "posix_time_timespec.hpp"

namespace posix {
/**
 * @namespace io
 * @brief Namspace for objects, envolved in the I/O process
 */
namespace io {
/**
 * @enum IOMode_t
 * @brief Read/Write mode
 */
enum class IOMode_t {
    /**
     * I/O operation returns like C single read/write, with the report of
     * actually received/transmitted bytes
	 */
	SINGLE,
    /**
	 * I/O operation make cycles, trying to receive/transmit requested data ammount
	 */
    LOOPED
};

/**
 * @enum AccessMode_t
 * @brief Node access mode
 */
enum class AccessMode_t {
	READ_ONLY = O_RDONLY,
	READ_WRITE = O_RDWR,
	WRITE_ONLY = O_WRONLY
};

POSIXCPP_FLAGS(Events, Events_t, short) {
	/**
	 * Data ready to read
	 */
	INPUT = POLLIN,
	    
	/**
	 * Data ready to write
	 */
	OUTPUT = POLLOUT,
	    
	/**
	 * High priority data ready
	 */
	URGENT = POLLPRI,
	    
	/**
	 * Error occured
	 */
	ERROR = POLLERR,
	
	HANGUP = POLLHUP,

	/**
	 * Invalid request
	 */
	INVALID = POLLNVAL
};

POSIXCPP_FLAGS(Mode, Mode_t, unsigned int) {
	NONE = 0,
	RWXU = S_IRWXU,
	RUSR = S_IRUSR,
	WUSR = S_IRUSR,
	XUSR = S_IXUSR,
	RWXG = S_IRWXG,
	RGRP = S_IRGRP,
	WGRP = S_IWGRP,
	XGRP = S_IXGRP,
	RWXO = S_IRWXO,
	ROTH = S_IROTH,
	WOTH = S_IWOTH,
	XOTH = S_IXOTH
};

enum class BlockingMode_t {
	BLOCK = 0,
	NON_BLOCK = O_NONBLOCK
};

/**
 * @interface IFile
 * @brief File operations interface
 */
class IHandle
{
public:
	
	virtual ~IHandle() = default;

	/**
	 * @brief Close file
	 */
	virtual void close() = 0;

	/**
	 * @brief Read data from file
	 * @param data Byte array to read to, can be reduced in size to actual read count
	 * @param mode Read mode ( single system call / looped untill the glorious end )
	 * @param timeout poll timeout for looped mode
	 */
	virtual void read(
	  posixcpp::ByteArray & data,
	  IOMode_t mode = IOMode_t::SINGLE,
	  time::Timespec timeout = -1 ) = 0;

	/**
	 * @brief Write data to file
	 * @param data Byte array to write from
	 * @param mode Write mode ( single system call / looped untill the glorious end )
	 * @param timeout poll timeout for looped mode
	 * @return actual read bytes
	 */
	virtual ssize_t write(
	  const posixcpp::ByteArray & data,
	  IOMode_t mode = IOMode_t::LOOPED,
	  time::Timespec timeout = -1 ) = 0;

	/**
	 * @brief Poll the device
	 * @param events
	 * @param timeout
	 * @return Occured events mask
	 */
	virtual Events poll(
	  Events events,
	  time::Timespec timeout = -1) = 0;
    
	virtual BlockingMode_t blockingMode() const = 0;
    
	virtual void setBlockingMode( BlockingMode_t newVal ) = 0;
};

/**
 * @brief ASCII control characters codes
 */
enum class ControlCharacters_t : uint8_t {
    /**
     * NULL
     */
    NUL = 0x00,
    /**
     * Start of header
     */
    SOH = 0x01,
    /**
     * Start of text
     */
    STX = 0x02,
    /**
     * End of text
     */
    ETX = 0x03,
    /**
     * End of transmission
     */
    EOT = 0x04,
    /**
     * End of query
     */
    ENQ = 0x05,
    /**
     * Acknowledge
     */
    ACK = 0x06,
    /**
     * Bell
     */
    BEL = 0x07,
    /**
     * Backspace
     */
    BS = 0x08,
    /**
     * Horizontal tab
     */
    HT = 0x09,
    /**
     * Line feed
     */
    LF = 0x0A,
    /**
     * Vertical tab
     */
    VT = 0x0B,
    /**
     * Form feed
     */
    FF = 0x0C,
    /**
     * Carriage return
     */
    CR = 0x0D,
    /**
     * Not acknowledge
     */
    NAK = 0x15
};

std::ostream &operator<<(std::ostream &stream, ControlCharacters_t cChar);

}
}

#endif // POSIX_IO_HPP_INCLUDED
