/*
    This file is part of Posixcpp library.

    Posixcpp library is free software: you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public License
    as published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    Posixcpp library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with Posixcpp library. If not, see
    <http://www.gnu.org/licenses/>.
 */

#ifndef POSIX_SOCKETS_SOCKET_HPP
#define	POSIX_SOCKETS_SOCKET_HPP

#include "posix_io_sockets.hpp"
#include "posix_io_handle.hpp"

#include <linux/if_packet.h>
#include <linux/if_ether.h>
#include <net/ethernet.h>
#include <sys/socket.h>
#include <bits/socket.h>
#include <arpa/inet.h>
#include <ostream>

namespace posix
{

namespace io
{

class Address
{
public:
    
    virtual ~Address() = default;
    
    virtual size_t length() const = 0;
    
    Family_t family ()
    {
        return Family_t(handle()->sa_family);
    }
    
    virtual sockaddr *handle() = 0;
    
protected:
    
    Address(sockaddr* addr, Family_t family = Family_t::UNSPECIFIED);
};

class PhisicalAddress final : public Address
{
public:
    
    PhisicalAddress(uint16_t ethernetProtocol = ETH_P_ALL);

    ~PhisicalAddress() noexcept = default;

    sockaddr* handle() override
    {
        return reinterpret_cast<sockaddr*>(&m_address);
    }

    void init(uint16_t ethernetProtocol = ETH_P_ALL);

    size_t length() const override;

    int ifIndex() const
    {
        return this->m_address.sll_ifindex;
    }

    void setIfIndex(int newVal)
    {
        this->m_address.sll_ifindex = newVal;
    }

    uint16_t protocol() const
    {
        return be16toh(this->m_address.sll_protocol);
    }

    void setProtocol(uint16_t newVal)
    {
        this->m_address.sll_protocol = htobe16(newVal);
    }
    
    unsigned char *address()
    {
        return m_address.sll_addr;
    }

private:
    sockaddr_ll m_address;
};


class InetAddress final : public Address
{
public:
    InetAddress(
      in_addr_t addr = INADDR_BROADCAST,
      in_port_t port = IPPORT_USERRESERVED+1 );

    InetAddress(const std::string & addr);
    
    ~InetAddress() noexcept = default;

    void init(
      in_addr_t addr = INADDR_BROADCAST,
      in_port_t port = IPPORT_USERRESERVED+1 );
    
    void setAddress(const std::string & addr);

    size_t length() const override;

    in_addr ip() const
    {
	    return this->m_address.sin_addr;
    }
    
    void setIp(in_addr_t newVal)
    {
	    this->m_address.sin_addr.s_addr = htobe32(newVal);
    }

    in_port_t port() const
    {
	    return be16toh(this->m_address.sin_port);
    }
    
    void setPort(in_port_t newVal)
    {
	    this->m_address.sin_port = htobe16(newVal);
    }
    
    sockaddr* handle() override
    {
        return reinterpret_cast<sockaddr*>(&m_address);
    }

private:

    sockaddr_in m_address;
};

std::ostream & operator << (std::ostream & stream, const InetAddress & addr);
    
class Socket : public io::Handle
{
public:
    
    Socket(
      Family_t domain,
      Type_t type,
      uint16_t protocol = 0);
    
    virtual ~Socket() noexcept = default;
    
    void create();
    
    void bind(const Address& addr);
    
    size_t send(
      const posixcpp::ByteArray & data,
      Flags_t flags = Flags_t::NONE);
    
    size_t sendTo(
      const posixcpp::ByteArray &data,
      const Address& addr,
      Flags_t flags = Flags_t::NONE);
    
    void listen();
    
    void accept(Socket & sock);
    
    void shutdown();

    Type_t type() const
    {
        return m_type;
    }

    void settype(Type_t newVal)
    {
        m_type = newVal;
    }
    
    void receive(
      posixcpp::ByteArray & data,
      Flags_t flags = Flags_t::NONE);

    size_t receiveBufferSize();
    
    void setReceiveBufferSize(size_t newVal);

    int protocol() const
    {
        return be16toh(m_protocol);
    }

    void setProtocol(int newVal)
    {
        m_protocol = htobe16(newVal);
    }

protected:
    
    Type_t m_type;
    
    Family_t m_domain;
    
    uint16_t m_protocol;
};

}
}

#endif	/* POSIX_SOCKETS_SOCKET_HPP */
