/*
    This file is part of Posixcpp library.

    Posixcpp library is free software: you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public License
    as published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    Posixcpp library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with Posixcpp library. If not, see
    <http://www.gnu.org/licenses/>.
*/

#ifndef POSIXCPP_HPP_INCLUDED
#define POSIXCPP_HPP_INCLUDED

#ifdef NO_NAMESPACE
#define POSIXCPP
#define NO_POSIXCPP
#else
#define POSIXCPP namespace posixcpp {
#define NO_POSIXCPP }
#endif

#if __cplusplus >= 201103L

#include <cstdint>
#include <system_error>
#define POSIXCPP_ENUM(name) enum class name
#define POSIXCPP_SIZED_ENUM(name, base) enum class name : base
#define SYSTEM_ERROR(text) \
throw std::system_error(errno, std::system_category(), text)

#else

#include <stdint.h>
#include <cstdlib>
#include <cstdio>
#include <cerrno>
#include <stdexcept>
#define POSIXCPP_ENUM(name) enum name
#define POSIXCPP_SIZED_ENUM(name, base) enum name
#define final
#define override

#define SYSTEM_ERROR(text) \
perror(text); \
throw std::runtime_error(text);

#endif

#include <stdexcept>
#include <string>
#include <ostream>

#define LINE_NUM(x) __STRING(x)

#ifdef POSIXCPP_USE_CONTRACTS

#define POSIX_CALL_ASSERT(condition) \
if ( !(condition) ) \
{ \
	SYSTEM_ERROR( \
        std::string("Call condition ").append( #condition ).append(" at "). \
        append(__FILE__).append(":").append(LINE_NUM( __LINE__ )).c_str()); \
}

#define POSIX_CALL_ASSERT_X(condition, message) \
if ( !(condition) ) \
{ \
	SYSTEM_ERROR( \
        std::string("Call condition ").append( #condition ).append(" at "). \
        append(__FILE__).append(":").append(LINE_NUM( __LINE__ )).append(" "). \
        append(message).c_str()); \
}

#define PRECONDITION_ASSERT( condition, message ) \
if ( !(condition) ) \
{ \
	throw std::logic_error(std::string("\n[PRECONDITION VIOLATION] ").append( \
		__PRETTY_FUNCTION__).append(": ").append(message).append("\n\tCondition \"" #condition \
		"\" not met in file " __FILE__ " line " LINE_NUM( __LINE__ )) ); \
}

/**
 * 
 */
#define POSTCONDITION_ASSERT( \
  condition, \
  message ) \
if (!(condition)) { \
	throw std::logic_error( std::string("\n[POSTCONDITION VIOLATION] ").append( \
		__PRETTY_FUNCTION__).append(": ").append(message).append("\n\tCondition \"" #condition \
		"\" not met in file " __FILE__ " line " LINE_NUM( __LINE__ )) ); \
}

#define INVARIANT_ASSERT( condition, message ) \
if ( !(condition) ) \
{ \
	throw std::logic_error( std::string("\n[INVARIANT VIOLATION] ").append( \
		__PRETTY_FUNCTION__).append(": ").append(message).append("\n\tCondition \"" #condition \
		"\" not met in file " __FILE__ " line " LINE_NUM( __LINE__ )) ); \
}

#else

#define POSIX_CALL_ASSERT(condition) \
if ( !(condition) ) \
{ \
	SYSTEM_ERROR(#condition); \
}

#define POSIX_CALL_ASSERT_X(condition, message) \
if ( !(condition) ) \
{ \
	SYSTEM_ERROR(message); \
}	
#define PRECONDITION_ASSERT( condition, message )
#define POSTCONDITION_ASSERT( condition, message )
#define INVARIANT_ASSERT( condition, message )
#endif

#endif // POSIXCPP_HPP_INCLUDED
