/*
    This file is part of Posixcpp library.

    Posixcpp library is free software: you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public License
    as published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    Posixcpp library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with Posixcpp library. If not, see
    <http://www.gnu.org/licenses/>.
*/

#include "posix_sockets_socket.hpp"

#include <string.h>

using namespace posix::io;

Address::Address( sockaddr* addr, Family_t family )
{
	addr->sa_family = int(family);
}

PhisicalAddress::PhisicalAddress( uint16_t ethernetProtocol ) :
Address((sockaddr*)(&m_address), Family_t::PACKET )
{
	this->init( ethernetProtocol );
}

void PhisicalAddress::init( uint16_t ethernetProtocol )
{
	this->m_address.sll_ifindex = 0;
	this->m_address.sll_protocol = htobe16(ethernetProtocol);
	this->m_address.sll_halen = ETH_ALEN;
	this->m_address.sll_pkttype = PACKET_BROADCAST;
	memset(this->m_address.sll_addr, 0xFF, ETH_ALEN);
}

size_t PhisicalAddress::length( ) const
{
	return sizeof(struct sockaddr_ll);
}

InetAddress::InetAddress( in_addr_t addr, in_port_t port ) :
Address((sockaddr*)(&m_address), Family_t::INET )
{
	this->init( addr, port );
}

InetAddress::InetAddress( const std::string& addr ) :
Address((sockaddr*)(&m_address), Family_t::INET )
{	
	setAddress(addr);
}

void
InetAddress::setAddress(const std::string& addr)
{
	size_t semicolonPos = addr.find(':');
	
	if (semicolonPos != std::string::npos)
	{
		in_port_t port;
		std::string str = addr.substr(0, semicolonPos);
		POSIX_CALL_ASSERT_X(inet_aton(str.c_str(), &m_address.sin_addr) != 0,
		    std::string("Invalid address ")+str );
		str = addr.substr(semicolonPos+1);
		port = std::atoi( str.c_str() );
		this->setPort(port);
	}
	else
		throw std::runtime_error("String should be <addr>:<port>");
}

void InetAddress::init( in_addr_t addr, in_port_t port )
{
	this->setPort(port);
	this->setIp(addr);
}

size_t InetAddress::length( ) const
{
	return sizeof(sockaddr_in);
}

Socket::Socket(
  Family_t domain,
  Type_t type,
  uint16_t protocol)
{
	this->m_domain = domain;
	this->m_type = type;
	this->m_protocol = htobe16(protocol);
}

void
Socket::create()
{
	POSIX_CALL_ASSERT( (this->m_descriptor =
	    socket(int(this->m_domain), int(this->m_type), m_protocol)) >= 0);
}

void
Socket::bind( const Address& addr )
{
	POSIX_CALL_ASSERT( ::bind( this->m_descriptor, const_cast<Address&>(addr).handle(),
	    addr.length() ) == 0);
}

size_t
Socket::send(
  const posixcpp::ByteArray& data,
  Flags_t flags)
{
	PRECONDITION_ASSERT(!data.isNull(), "data is empty");
	PRECONDITION_ASSERT(data.size() > 0, "data is empty");
	PRECONDITION_ASSERT(m_descriptor > 0, "socket isn't open");
	
	// Количество записанных байт
	ssize_t count;
	
	if (m_descriptor <= 0)
		throw std::runtime_error("socket isn't opened");
	
	POSIX_CALL_ASSERT((count = ::send(this->m_descriptor, data.data(),
	    data.size(), int(flags))) > -1);
	
	POSTCONDITION_ASSERT(count <= ssize_t(data.size()), "Invalid result");
	if (count < 0)
		count = 0;
	
	return size_t(count);
}

size_t
Socket::sendTo(
  const posixcpp::ByteArray& data,
  const Address& addr,
  Flags_t flags)
{
	PRECONDITION_ASSERT(!data.isNull(), "data is empty");
	PRECONDITION_ASSERT(data.size() > 0, "data is empty");
	PRECONDITION_ASSERT(m_descriptor > 0, "socket isn't open");
	
	// Количество записанных байт
	ssize_t count;
	
	if (m_descriptor <= 0)
		throw std::runtime_error("socket isn't opened");
	
	POSIX_CALL_ASSERT((count = ::sendto( this->m_descriptor, data.data(),
	    data.size(), int(flags), const_cast<Address&>(addr).handle(),
	    addr.length())) > -1);
	
	POSTCONDITION_ASSERT(count <= ssize_t(data.size()), "Invalid result");
	if (count < 0)
		count = 0;
	
	return count;
}

void
Socket::receive(
  posixcpp::ByteArray& data,
  Flags_t flags)
{
	PRECONDITION_ASSERT(!data.isNull(), "data is empty");
	PRECONDITION_ASSERT(data.size() > 0, "data is empty");
	PRECONDITION_ASSERT(m_descriptor > 0, "socket isn't open");
	
	ssize_t count;
	
	POSIX_CALL_ASSERT((count = ::recv(this->m_descriptor, data.data(),
	    data.size(), int(flags))) > -1);
	
	POSTCONDITION_ASSERT(count <= ssize_t(data.size()), "Invalid result");
	if (count < 0)
		count = 0;
	
	data.resize(count);
}

void Socket::listen()
{
	POSIX_CALL_ASSERT(::listen(this->m_descriptor, 0) == 0);
}

void Socket::accept(Socket& sock)
{
	int descriptor;

	POSIX_CALL_ASSERT((descriptor=::accept(this->m_descriptor, NULL, 0)) > 0);
	
	sock.setBlockingMode(this->blockingMode());
	sock.setProtocol(this->protocol());
	sock.setReadTries(this->readTries());
	sock.settype(this->type());
	sock.setDescriptor( descriptor );
}

namespace posix
{
namespace io
{
namespace sockets
{
std::ostream & operator << (std::ostream & stream, const InetAddress & addr)
{
	std::string str(inet_ntoa(addr.ip()));
	stream << str << ':' << addr.port();
	
	return stream;
}
}
}
}
